@@ -45,6 +45,7 @@ import org.xnio.ssl.JsseSslUtils;
 import javax.net.ssl.KeyManager;
 import javax.net.ssl.SSLContext;
 import javax.net.ssl.TrustManager;
+import java.io.IOException;
 import java.net.Inet4Address;
 import java.net.InetSocketAddress;
 import java.net.SocketAddress;
@@ -168,7 +169,7 @@ public final class Undertow {
                     AcceptingChannel<? extends StreamConnection> server = worker.createStreamConnectionServer(new InetSocketAddress(Inet4Address.getByName(listener.host), listener.port), acceptListener, socketOptionsWithOverrides);
                     server.resumeAccepts();
                     channels.add(server);
-                    listenerInfo.add(new ListenerInfo("ajp", server.getLocalAddress(), null, openListener));
+                    listenerInfo.add(new ListenerInfo("ajp", server.getLocalAddress(), openListener, null, server));
                 } else {
                     OptionMap undertowOptions = OptionMap.builder().set(UndertowOptions.BUFFER_PIPELINED_DATA, true).addAll(serverOptions).getMap();
                     boolean http2 = serverOptions.get(UndertowOptions.ENABLE_HTTP2, false);
@@ -191,7 +192,7 @@ public final class Undertow {
                         AcceptingChannel<? extends StreamConnection> server = worker.createStreamConnectionServer(new InetSocketAddress(Inet4Address.getByName(listener.host), listener.port), acceptListener, socketOptionsWithOverrides);
                         server.resumeAccepts();
                         channels.add(server);
-                        listenerInfo.add(new ListenerInfo("http", server.getLocalAddress(), null, openListener));
+                        listenerInfo.add(new ListenerInfo("http", server.getLocalAddress(), openListener, null, server));
                     } else if (listener.type == ListenerType.HTTPS) {
                         OpenListener openListener;
 
@@ -235,13 +236,16 @@ public final class Undertow {
 
                         sslServer.resumeAccepts();
                         channels.add(sslServer);
-                        listenerInfo.add(new ListenerInfo("https", sslServer.getLocalAddress(), listener.sslContext, openListener));
+                        listenerInfo.add(new ListenerInfo("https", sslServer.getLocalAddress(), openListener, xnioSsl, sslServer));
                     }
                 }
 
             }
 
         } catch (Exception e) {
+            if(internalWorker && worker != null) {
+                worker.shutdownNow();
+            }
             throw new RuntimeException(e);
         }
     }
@@ -259,7 +263,12 @@ public final class Undertow {
          * Only shutdown the worker if it was created during start()
          */
         if (internalWorker && worker != null) {
-            worker.shutdownNow();
+            worker.shutdown();
+            try {
+                worker.awaitTermination();
+            } catch (InterruptedException e) {
+                throw new RuntimeException(e);
+            }
             worker = null;
         }
         xnio = null;
@@ -427,7 +436,7 @@ public final class Undertow {
                 //use 16k buffers for best performance
                 //as 16k is generally the max amount of data that can be sent in a single write() call
                 directBuffers = true;
-                bufferSize = 1024 * 16;
+                bufferSize = 1024 * 16 - 20; //the 20 is to allow some space for protocol headers, see UNDERTOW-1209
             }
 
         }
@@ -565,14 +574,16 @@ public final class Undertow {
 
         private final String protcol;
         private final SocketAddress address;
-        private final SSLContext sslContext;
         private final OpenListener openListener;
+        private final UndertowXnioSsl ssl;
+        private final AcceptingChannel<? extends StreamConnection> channel;
 
-        public ListenerInfo(String protcol, SocketAddress address, SSLContext sslContext, OpenListener openListener) {
+        public ListenerInfo(String protcol, SocketAddress address, OpenListener openListener, UndertowXnioSsl ssl, AcceptingChannel<? extends StreamConnection> channel) {
             this.protcol = protcol;
             this.address = address;
-            this.sslContext = sslContext;
             this.openListener = openListener;
+            this.ssl = ssl;
+            this.channel = channel;
         }
 
         public String getProtcol() {
@@ -584,19 +595,37 @@ public final class Undertow {
         }
 
         public SSLContext getSslContext() {
-            return sslContext;
+            if(ssl == null) {
+                return null;
+            }
+            return ssl.getSslContext();
+        }
+
+        public void setSslContext(SSLContext sslContext) {
+            if(ssl != null) {
+                //just ignore it if this is not a SSL listener
+                ssl.updateSSLContext(sslContext);
+            }
         }
 
         public ConnectorStatistics getConnectorStatistics() {
             return openListener.getConnectorStatistics();
         }
 
+        public <T> void setSocketOption(Option<T>option, T value) throws IOException {
+            channel.setOption(option, value);
+        }
+
+        public void setServerOptions(OptionMap options) {
+            openListener.setUndertowOptions(options);
+        }
+
         @Override
         public String toString() {
             return "ListenerInfo{" +
                     "protcol='" + protcol + '\'' +
                     ", address=" + address +
-                    ", sslContext=" + sslContext +
+                    ", sslContext=" + getSslContext() +
                     '}';
         }
     }
