@@ -30,6 +30,7 @@ import io.undertow.server.Connectors;
 import io.undertow.server.HttpHandler;
 import io.undertow.server.HttpServerExchange;
 import io.undertow.server.protocol.http.HttpAttachments;
+import io.undertow.server.protocol.http.HttpContinue;
 import io.undertow.util.HeaderMap;
 import io.undertow.util.HeaderValues;
 import io.undertow.util.Headers;
@@ -73,6 +74,7 @@ public class Http2ReceiveListener implements ChannelListener<Http2Channel> {
     private final boolean allowEncodingSlash;
     private final int bufferSize;
     private final int maxParameters;
+    private final boolean recordRequestStartTime;
 
 
 
@@ -95,6 +97,7 @@ public class Http2ReceiveListener implements ChannelListener<Http2Channel> {
         this.allowEncodingSlash = undertowOptions.get(UndertowOptions.ALLOW_ENCODED_SLASH, false);
         this.decode = undertowOptions.get(UndertowOptions.DECODE_URL, true);
         this.maxParameters = undertowOptions.get(UndertowOptions.MAX_PARAMETERS, UndertowOptions.DEFAULT_MAX_PARAMETERS);
+        this.recordRequestStartTime = undertowOptions.get(UndertowOptions.RECORD_REQUEST_START_TIME, false);
         if (undertowOptions.get(UndertowOptions.DECODE_URL, true)) {
             this.encoding = undertowOptions.get(UndertowOptions.URL_CHARSET, StandardCharsets.UTF_8.name());
         } else {
@@ -173,14 +176,9 @@ public class Http2ReceiveListener implements ChannelListener<Http2Channel> {
             channel.sendGoAway(Http2Channel.ERROR_PROTOCOL_ERROR);
             return;
         }
-        try {
-            Connectors.setExchangeRequestPath(exchange, path, encoding, decode, allowEncodingSlash, decodeBuffer, maxParameters);
-        } catch (ParameterLimitException e) {
-            //this can happen if max parameters is exceeded
-            UndertowLogger.REQUEST_IO_LOGGER.debug("Failed to set request path", e);
-            exchange.setStatusCode(StatusCodes.BAD_REQUEST);
-            exchange.endExchange();
-            return;
+
+        if (recordRequestStartTime) {
+            Connectors.setRequestStartTime(exchange);
         }
         SSLSession session = channel.getSslSession();
         if(session != null) {
@@ -206,6 +204,16 @@ public class Http2ReceiveListener implements ChannelListener<Http2Channel> {
             connectorStatistics.setup(exchange);
         }
 
+        try {
+            Connectors.setExchangeRequestPath(exchange, path, encoding, decode, allowEncodingSlash, decodeBuffer, maxParameters);
+        } catch (ParameterLimitException e) {
+            //this can happen if max parameters is exceeded
+            UndertowLogger.REQUEST_IO_LOGGER.debug("Failed to set request path", e);
+            exchange.setStatusCode(StatusCodes.BAD_REQUEST);
+            exchange.endExchange();
+            return;
+        }
+
         //TODO: we should never actually put these into the map in the first place
         exchange.getRequestHeaders().remove(AUTHORITY);
         exchange.getRequestHeaders().remove(PATH);
@@ -234,9 +242,13 @@ public class Http2ReceiveListener implements ChannelListener<Http2Channel> {
             requestHeaders.putAll(hv.getHeaderName(), hv);
         }
         final HttpServerExchange exchange = new HttpServerExchange(connection, requestHeaders, sink.getHeaders(), maxEntitySize);
+        if(initial.getRequestHeaders().contains(Headers.EXPECT)) {
+            HttpContinue.markContinueResponseSent(exchange);
+        }
         if(initial.getAttachment(HttpAttachments.REQUEST_TRAILERS) != null) {
             exchange.putAttachment(HttpAttachments.REQUEST_TRAILERS, initial.getAttachment(HttpAttachments.REQUEST_TRAILERS));
         }
+        Connectors.setRequestStartTime(initial, exchange);
         connection.setExchange(exchange);
         exchange.setRequestScheme(initial.getRequestScheme());
         exchange.setProtocol(initial.getProtocol());
@@ -291,7 +303,7 @@ public class Http2ReceiveListener implements ChannelListener<Http2Channel> {
 
         // if CONNECT type is used, then we expect :method and :authority to be present only;
         // :scheme and :path must not be present
-        if (headers.get(METHOD).contains(Methods.CONNECT)) {
+        if (headers.get(METHOD).contains(Methods.CONNECT_STRING)) {
             if (headers.contains(SCHEME) || headers.contains(PATH) || headers.count(AUTHORITY) != 1) {
                 return false;
             }
