@@ -51,6 +51,8 @@ import io.undertow.util.HttpString;
 import io.undertow.util.RedirectBuilder;
 import io.undertow.util.StatusCodes;
 
+import static io.undertow.util.URLUtils.isAbsoluteUrl;
+
 
 /**
  * @author Stuart Douglas
@@ -184,7 +186,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         resetBuffer();
         setStatus(StatusCodes.FOUND);
         String realPath;
-        if (location.contains("://")) {//absolute url
+        if (isAbsoluteUrl(location)) {//absolute url
             exchange.getResponseHeaders().put(Headers.LOCATION, location);
         } else {
             if (location.startsWith("/")) {
@@ -318,7 +320,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     @Override
     public String getCharacterEncoding() {
         if (charset == null) {
-            return servletContext.getDeployment().getDeploymentInfo().getDefaultEncoding();
+            return servletContext.getDeployment().getDefaultCharset().name();
         }
         return charset;
     }
@@ -390,7 +392,11 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         if (insideInclude || responseStarted()) {
             return;
         }
-        exchange.getResponseHeaders().put(Headers.CONTENT_LENGTH, Integer.toString(len));
+        if(len >= 0) {
+            exchange.getResponseHeaders().put(Headers.CONTENT_LENGTH, Integer.toString(len));
+        } else {
+            exchange.getResponseHeaders().remove(Headers.CONTENT_LENGTH);
+        }
         this.contentLength = (long) len;
     }
 
@@ -399,7 +405,11 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         if (insideInclude || responseStarted()) {
             return;
         }
-        exchange.getResponseHeaders().put(Headers.CONTENT_LENGTH, Long.toString(len));
+        if(len >= 0) {
+            exchange.getResponseHeaders().put(Headers.CONTENT_LENGTH, Long.toString(len));
+        } else {
+            exchange.getResponseHeaders().remove(Headers.CONTENT_LENGTH);
+        }
         this.contentLength = len;
     }
 
